import { join } from "./String.js";
import { defaultArg, some, value as value_1, Option } from "./Option.js";
import { IComparer, IEqualityComparer, disposeSafe, isArrayLike, defaultOf, getEnumerator, toIterator, IEnumerator, compare, structuralHash, equals } from "./Util.js";
import { int32 } from "./Int32.js";
import { Record } from "./Types.js";
import { class_type, record_type, option_type, TypeInfo } from "./Reflection.js";
import { SR_inputSequenceTooLong, SR_inputSequenceEmpty, SR_inputMustBeNonNegative, SR_notEnoughElements, SR_differentLengths, SR_keyNotFoundAlt, SR_indexOutOfBounds, SR_inputWasEmpty } from "./Global.js";
import { transpose as transpose_1, splitInto as splitInto_1, windowed as windowed_1, pairwise as pairwise_1, chunkBySize as chunkBySize_1, map as map_1, permute as permute_1, tryFindIndexBack as tryFindIndexBack_1, tryFindBack as tryFindBack_1, scanBack as scanBack_1, foldBack2 as foldBack2_1, foldBack as foldBack_1, fill } from "./Array.js";

export class FSharpList<T> extends Record {
    head: T;
    tail: Option<FSharpList<T>>;
    constructor(head: T, tail: Option<FSharpList<T>>) {
        super();
        this.head = head;
        this.tail = tail;
    }
    toString(): string {
        const xs: FSharpList<T> = this;
        return ("[" + join("; ", xs)) + "]";
    }
    Equals(other: any): boolean {
        const xs: FSharpList<T> = this;
        if (xs === other) {
            return true;
        }
        else {
            const loop: (arg0: FSharpList<T>, arg1: FSharpList<T>) => boolean = (xs_1_mut: FSharpList<T>, ys_1_mut: FSharpList<T>): boolean => {
                loop:
                while (true) {
                    const xs_1: FSharpList<T> = xs_1_mut, ys_1: FSharpList<T> = ys_1_mut;
                    const matchValue: Option<FSharpList<T>> = xs_1.tail;
                    const matchValue_1: Option<FSharpList<T>> = ys_1.tail;
                    if (matchValue != null) {
                        if (matchValue_1 != null) {
                            const xt: FSharpList<T> = value_1(matchValue);
                            const yt: FSharpList<T> = value_1(matchValue_1);
                            if (equals(xs_1.head, ys_1.head)) {
                                xs_1_mut = xt;
                                ys_1_mut = yt;
                                continue loop;
                            }
                            else {
                                return false;
                            }
                        }
                        else {
                            return false;
                        }
                    }
                    else if (matchValue_1 != null) {
                        return false;
                    }
                    else {
                        return true;
                    }
                    break;
                }
            };
            return loop(xs, other);
        }
    }
    GetHashCode(): int32 {
        const xs: FSharpList<T> = this;
        const loop: (arg0: int32, arg1: int32, arg2: FSharpList<T>) => int32 = (i_mut: int32, h_mut: int32, xs_1_mut: FSharpList<T>): int32 => {
            loop:
            while (true) {
                const i: int32 = i_mut, h: int32 = h_mut, xs_1: FSharpList<T> = xs_1_mut;
                const matchValue: Option<FSharpList<T>> = xs_1.tail;
                if (matchValue != null) {
                    const t: FSharpList<T> = value_1(matchValue);
                    if (i > 18) {
                        return h | 0;
                    }
                    else {
                        i_mut = (i + 1);
                        h_mut = (((h << 1) + structuralHash(xs_1.head)) + (631 * i));
                        xs_1_mut = t;
                        continue loop;
                    }
                }
                else {
                    return h | 0;
                }
                break;
            }
        };
        return loop(0, 0, xs) | 0;
    }
    toJSON(): any {
        const this$: FSharpList<T> = this;
        return Array.from(this$);
    }
    CompareTo(other: any): int32 {
        const xs: FSharpList<T> = this;
        const loop: (arg0: FSharpList<T>, arg1: FSharpList<T>) => int32 = (xs_1_mut: FSharpList<T>, ys_1_mut: FSharpList<T>): int32 => {
            loop:
            while (true) {
                const xs_1: FSharpList<T> = xs_1_mut, ys_1: FSharpList<T> = ys_1_mut;
                const matchValue: Option<FSharpList<T>> = xs_1.tail;
                const matchValue_1: Option<FSharpList<T>> = ys_1.tail;
                if (matchValue != null) {
                    if (matchValue_1 != null) {
                        const xt: FSharpList<T> = value_1(matchValue);
                        const yt: FSharpList<T> = value_1(matchValue_1);
                        const c: int32 = compare(xs_1.head, ys_1.head) | 0;
                        if (c === 0) {
                            xs_1_mut = xt;
                            ys_1_mut = yt;
                            continue loop;
                        }
                        else {
                            return c | 0;
                        }
                    }
                    else {
                        return 1;
                    }
                }
                else if (matchValue_1 != null) {
                    return -1;
                }
                else {
                    return 0;
                }
                break;
            }
        };
        return loop(xs, other) | 0;
    }
    GetEnumerator(): IEnumerator<T> {
        const xs: FSharpList<T> = this;
        return ListEnumerator$1_$ctor_3002E699(xs);
    }
    [Symbol.iterator]() {
        return toIterator(this.GetEnumerator());
    }
    "System.Collections.IEnumerable.GetEnumerator"(): IEnumerator<any> {
        const xs: FSharpList<T> = this;
        return getEnumerator(xs);
    }
}

export function FSharpList$reflection(gen0: TypeInfo): TypeInfo {
    return record_type("ListModule.FSharpList", [gen0], FSharpList, () => [["head", gen0], ["tail", option_type(FSharpList$reflection(gen0))]]);
}

export class ListEnumerator$1<T> {
    xs: FSharpList<T>;
    it: FSharpList<T>;
    current: T;
    constructor(xs: FSharpList<T>) {
        this.xs = xs;
        this.it = this.xs;
        this.current = defaultOf();
    }
    "System.Collections.Generic.IEnumerator`1.get_Current"(): T {
        const _: ListEnumerator$1<T> = this;
        return _.current;
    }
    "System.Collections.IEnumerator.get_Current"(): any {
        const _: ListEnumerator$1<T> = this;
        return _.current;
    }
    "System.Collections.IEnumerator.MoveNext"(): boolean {
        const _: ListEnumerator$1<T> = this;
        const matchValue: Option<FSharpList<T>> = _.it.tail;
        if (matchValue != null) {
            const t: FSharpList<T> = value_1(matchValue);
            _.current = _.it.head;
            _.it = t;
            return true;
        }
        else {
            return false;
        }
    }
    "System.Collections.IEnumerator.Reset"(): void {
        const _: ListEnumerator$1<T> = this;
        _.it = _.xs;
        _.current = defaultOf();
    }
    Dispose(): void {
    }
}

export function ListEnumerator$1$reflection(gen0: TypeInfo): TypeInfo {
    return class_type("ListModule.ListEnumerator`1", [gen0], ListEnumerator$1);
}

export function ListEnumerator$1_$ctor_3002E699<T>(xs: FSharpList<T>): ListEnumerator$1<T> {
    return new ListEnumerator$1(xs);
}

export function FSharpList_get_Empty<T>(): FSharpList<T> {
    return new FSharpList(defaultOf(), void 0);
}

export function FSharpList_Cons_305B8EAC<T>(x: T, xs: FSharpList<T>): FSharpList<T> {
    return new FSharpList(x, xs);
}

export function FSharpList__get_IsEmpty<T>(xs: FSharpList<T>): boolean {
    return xs.tail == null;
}

export function FSharpList__get_Length<T>(xs: FSharpList<T>): int32 {
    const loop: (arg0: int32, arg1: FSharpList<T>) => int32 = (i_mut: int32, xs_1_mut: FSharpList<T>): int32 => {
        loop:
        while (true) {
            const i: int32 = i_mut, xs_1: FSharpList<T> = xs_1_mut;
            const matchValue: Option<FSharpList<T>> = xs_1.tail;
            if (matchValue != null) {
                i_mut = (i + 1);
                xs_1_mut = value_1(matchValue);
                continue loop;
            }
            else {
                return i | 0;
            }
            break;
        }
    };
    return loop(0, xs) | 0;
}

export function FSharpList__get_Head<T>(xs: FSharpList<T>): T {
    const matchValue: Option<FSharpList<T>> = xs.tail;
    if (matchValue != null) {
        return xs.head;
    }
    else {
        throw new Error((SR_inputWasEmpty + "\\nParameter name: ") + "list");
    }
}

export function FSharpList__get_Tail<T>(xs: FSharpList<T>): FSharpList<T> {
    const matchValue: Option<FSharpList<T>> = xs.tail;
    if (matchValue != null) {
        return value_1(matchValue);
    }
    else {
        throw new Error((SR_inputWasEmpty + "\\nParameter name: ") + "list");
    }
}

export function FSharpList__get_Item_Z524259A4<T>(xs: FSharpList<T>, index: int32): T {
    const loop: (arg0: int32, arg1: FSharpList<T>) => T = (i_mut: int32, xs_1_mut: FSharpList<T>): T => {
        loop:
        while (true) {
            const i: int32 = i_mut, xs_1: FSharpList<T> = xs_1_mut;
            const matchValue: Option<FSharpList<T>> = xs_1.tail;
            if (matchValue != null) {
                if (i === index) {
                    return xs_1.head;
                }
                else {
                    i_mut = (i + 1);
                    xs_1_mut = value_1(matchValue);
                    continue loop;
                }
            }
            else {
                throw new Error((SR_indexOutOfBounds + "\\nParameter name: ") + "index");
            }
            break;
        }
    };
    return loop(0, xs);
}

export function indexNotFound<$a>(): $a {
    throw new Error(SR_keyNotFoundAlt);
}

export function empty<$a>(): FSharpList<$a> {
    return FSharpList_get_Empty();
}

export function cons<T>(x: T, xs: FSharpList<T>): FSharpList<T> {
    return FSharpList_Cons_305B8EAC(x, xs);
}

export function singleton<$a>(x: $a): FSharpList<$a> {
    return FSharpList_Cons_305B8EAC(x, FSharpList_get_Empty());
}

export function isEmpty<T>(xs: FSharpList<T>): boolean {
    return FSharpList__get_IsEmpty(xs);
}

export function length<T>(xs: FSharpList<T>): int32 {
    return FSharpList__get_Length(xs);
}

export function head<T>(xs: FSharpList<T>): T {
    return FSharpList__get_Head(xs);
}

export function tryHead<T>(xs: FSharpList<T>): Option<T> {
    if (FSharpList__get_IsEmpty(xs)) {
        return void 0;
    }
    else {
        return some(FSharpList__get_Head(xs));
    }
}

export function tail<T>(xs: FSharpList<T>): FSharpList<T> {
    return FSharpList__get_Tail(xs);
}

export function tryLast<T>(xs_mut: FSharpList<T>): Option<T> {
    tryLast:
    while (true) {
        const xs: FSharpList<T> = xs_mut;
        if (FSharpList__get_IsEmpty(xs)) {
            return void 0;
        }
        else {
            const t: FSharpList<T> = FSharpList__get_Tail(xs);
            if (FSharpList__get_IsEmpty(t)) {
                return some(FSharpList__get_Head(xs));
            }
            else {
                xs_mut = t;
                continue tryLast;
            }
        }
        break;
    }
}

export function last<T>(xs: FSharpList<T>): T {
    const matchValue: Option<T> = tryLast(xs);
    if (matchValue == null) {
        throw new Error(SR_inputWasEmpty);
    }
    else {
        return value_1(matchValue);
    }
}

export function compareWith<T>(comparer: (arg0: T, arg1: T) => int32, xs: FSharpList<T>, ys: FSharpList<T>): int32 {
    const loop: (arg0: FSharpList<T>, arg1: FSharpList<T>) => int32 = (xs_1_mut: FSharpList<T>, ys_1_mut: FSharpList<T>): int32 => {
        loop:
        while (true) {
            const xs_1: FSharpList<T> = xs_1_mut, ys_1: FSharpList<T> = ys_1_mut;
            const matchValue: boolean = FSharpList__get_IsEmpty(xs_1);
            const matchValue_1: boolean = FSharpList__get_IsEmpty(ys_1);
            if (matchValue) {
                if (matchValue_1) {
                    return 0;
                }
                else {
                    return -1;
                }
            }
            else if (matchValue_1) {
                return 1;
            }
            else {
                const c: int32 = comparer(FSharpList__get_Head(xs_1), FSharpList__get_Head(ys_1)) | 0;
                if (c === 0) {
                    xs_1_mut = FSharpList__get_Tail(xs_1);
                    ys_1_mut = FSharpList__get_Tail(ys_1);
                    continue loop;
                }
                else {
                    return c | 0;
                }
            }
            break;
        }
    };
    return loop(xs, ys) | 0;
}

export function toArray<T>(xs: FSharpList<T>): T[] {
    const len: int32 = FSharpList__get_Length(xs) | 0;
    const res: T[] = fill(new Array(len), 0, len, null);
    const loop: (arg0: int32, arg1: FSharpList<T>) => void = (i_mut: int32, xs_1_mut: FSharpList<T>): void => {
        loop:
        while (true) {
            const i: int32 = i_mut, xs_1: FSharpList<T> = xs_1_mut;
            if (!FSharpList__get_IsEmpty(xs_1)) {
                res[i] = FSharpList__get_Head(xs_1);
                i_mut = (i + 1);
                xs_1_mut = FSharpList__get_Tail(xs_1);
                continue loop;
            }
            break;
        }
    };
    loop(0, xs);
    return res;
}

export function fold<State, T>(folder: (arg0: State, arg1: T) => State, state: State, xs: FSharpList<T>): State {
    let acc: State = state;
    let xs_1: FSharpList<T> = xs;
    while (!FSharpList__get_IsEmpty(xs_1)) {
        acc = folder(acc, FSharpList__get_Head(xs_1));
        xs_1 = FSharpList__get_Tail(xs_1);
    }
    return acc;
}

export function reverse<T>(xs: FSharpList<T>): FSharpList<T> {
    return fold((acc: FSharpList<T>, x: T): FSharpList<T> => FSharpList_Cons_305B8EAC(x, acc), FSharpList_get_Empty(), xs);
}

export function foldBack<T, State>(folder: (arg0: T, arg1: State) => State, xs: FSharpList<T>, state: State): State {
    return foldBack_1(folder, toArray(xs), state);
}

export function foldIndexed<State, T>(folder: (arg0: int32, arg1: State, arg2: T) => State, state: State, xs: FSharpList<T>): State {
    const loop: (arg0: int32, arg1: State, arg2: FSharpList<T>) => State = (i_mut: int32, acc_mut: State, xs_1_mut: FSharpList<T>): State => {
        loop:
        while (true) {
            const i: int32 = i_mut, acc: State = acc_mut, xs_1: FSharpList<T> = xs_1_mut;
            if (FSharpList__get_IsEmpty(xs_1)) {
                return acc;
            }
            else {
                i_mut = (i + 1);
                acc_mut = folder(i, acc, FSharpList__get_Head(xs_1));
                xs_1_mut = FSharpList__get_Tail(xs_1);
                continue loop;
            }
            break;
        }
    };
    return loop(0, state, xs);
}

export function fold2<State, T1, T2>(folder: (arg0: State, arg1: T1, arg2: T2) => State, state: State, xs: FSharpList<T1>, ys: FSharpList<T2>): State {
    let acc: State = state;
    let xs_1: FSharpList<T1> = xs;
    let ys_1: FSharpList<T2> = ys;
    while ((!FSharpList__get_IsEmpty(xs_1)) && (!FSharpList__get_IsEmpty(ys_1))) {
        acc = folder(acc, FSharpList__get_Head(xs_1), FSharpList__get_Head(ys_1));
        xs_1 = FSharpList__get_Tail(xs_1);
        ys_1 = FSharpList__get_Tail(ys_1);
    }
    return acc;
}

export function foldBack2<T1, T2, State>(folder: (arg0: T1, arg1: T2, arg2: State) => State, xs: FSharpList<T1>, ys: FSharpList<T2>, state: State): State {
    return foldBack2_1(folder, toArray(xs), toArray(ys), state);
}

export function unfold<State, T>(gen: (arg0: State) => Option<[T, State]>, state: State): FSharpList<T> {
    const loop: (arg0: State, arg1: FSharpList<T>) => FSharpList<T> = (acc_mut: State, node_mut: FSharpList<T>): FSharpList<T> => {
        let t: FSharpList<T>;
        loop:
        while (true) {
            const acc: State = acc_mut, node: FSharpList<T> = node_mut;
            const matchValue: Option<[T, State]> = gen(acc);
            if (matchValue != null) {
                acc_mut = value_1(matchValue)[1];
                node_mut = ((t = (new FSharpList(value_1(matchValue)[0], void 0)), (node.tail = t, t)));
                continue loop;
            }
            else {
                return node;
            }
            break;
        }
    };
    const root: FSharpList<T> = FSharpList_get_Empty();
    const node_1: FSharpList<T> = loop(state, root);
    const t_2: FSharpList<T> = FSharpList_get_Empty();
    node_1.tail = t_2;
    return FSharpList__get_Tail(root);
}

export function iterate<$a>(action: (arg0: $a) => void, xs: FSharpList<$a>): void {
    fold((unitVar: void, x: $a): void => {
        action(x);
    }, void 0, xs);
}

export function iterate2<$a, $b>(action: (arg0: $a, arg1: $b) => void, xs: FSharpList<$a>, ys: FSharpList<$b>): void {
    fold2((unitVar: void, x: $a, y: $b): void => {
        action(x, y);
    }, void 0, xs, ys);
}

export function iterateIndexed<$a>(action: (arg0: int32, arg1: $a) => void, xs: FSharpList<$a>): void {
    fold((i: int32, x: $a): int32 => {
        action(i, x);
        return (i + 1) | 0;
    }, 0, xs);
}

export function iterateIndexed2<$a, $b>(action: (arg0: int32, arg1: $a, arg2: $b) => void, xs: FSharpList<$a>, ys: FSharpList<$b>): void {
    fold2((i: int32, x: $a, y: $b): int32 => {
        action(i, x, y);
        return (i + 1) | 0;
    }, 0, xs, ys);
}

export function toSeq<T>(xs: FSharpList<T>): Iterable<T> {
    return xs;
}

export function ofArrayWithTail<T>(xs: T[], tail_1: FSharpList<T>): FSharpList<T> {
    let res: FSharpList<T> = tail_1;
    for (let i: int32 = xs.length - 1; i >= 0; i--) {
        res = FSharpList_Cons_305B8EAC(xs[i], res);
    }
    return res;
}

export function ofArray<T>(xs: T[]): FSharpList<T> {
    return ofArrayWithTail(xs, FSharpList_get_Empty());
}

export function ofSeq<T>(xs: Iterable<T>): FSharpList<T> {
    let xs_3: FSharpList<T>, t: FSharpList<T>;
    if (isArrayLike(xs)) {
        return ofArray(xs);
    }
    else if (xs instanceof FSharpList) {
        return xs;
    }
    else {
        const root: FSharpList<T> = FSharpList_get_Empty();
        let node: FSharpList<T> = root;
        const enumerator: IEnumerator<T> = getEnumerator(xs);
        try {
            while (enumerator["System.Collections.IEnumerator.MoveNext"]()) {
                const x: T = enumerator["System.Collections.Generic.IEnumerator`1.get_Current"]();
                node = ((xs_3 = node, (t = (new FSharpList(x, void 0)), (xs_3.tail = t, t))));
            }
        }
        finally {
            disposeSafe(enumerator);
        }
        const xs_5: FSharpList<T> = node;
        const t_2: FSharpList<T> = FSharpList_get_Empty();
        xs_5.tail = t_2;
        return FSharpList__get_Tail(root);
    }
}

export function concat<T>(lists: Iterable<FSharpList<T>>): FSharpList<T> {
    const root: FSharpList<T> = FSharpList_get_Empty();
    let node: FSharpList<T> = root;
    const action: (arg0: FSharpList<T>) => void = (xs: FSharpList<T>): void => {
        node = fold((acc: FSharpList<T>, x: T): FSharpList<T> => {
            const t: FSharpList<T> = new FSharpList(x, void 0);
            acc.tail = t;
            return t;
        }, node, xs);
    };
    if (isArrayLike(lists)) {
        lists.forEach(action);
    }
    else if (lists instanceof FSharpList) {
        iterate(action, lists);
    }
    else {
        const enumerator: IEnumerator<FSharpList<T>> = getEnumerator(lists);
        try {
            while (enumerator["System.Collections.IEnumerator.MoveNext"]()) {
                action(enumerator["System.Collections.Generic.IEnumerator`1.get_Current"]());
            }
        }
        finally {
            disposeSafe(enumerator);
        }
    }
    const xs_6: FSharpList<T> = node;
    const t_2: FSharpList<T> = FSharpList_get_Empty();
    xs_6.tail = t_2;
    return FSharpList__get_Tail(root);
}

export function scan<State, T>(folder: (arg0: State, arg1: T) => State, state: State, xs: FSharpList<T>): FSharpList<State> {
    let xs_4: FSharpList<State>, t_2: FSharpList<State>;
    const root: FSharpList<State> = FSharpList_get_Empty();
    let node: FSharpList<State>;
    const t: FSharpList<State> = new FSharpList(state, void 0);
    root.tail = t;
    node = t;
    let acc: State = state;
    let xs_3: FSharpList<T> = xs;
    while (!FSharpList__get_IsEmpty(xs_3)) {
        acc = folder(acc, FSharpList__get_Head(xs_3));
        node = ((xs_4 = node, (t_2 = (new FSharpList(acc, void 0)), (xs_4.tail = t_2, t_2))));
        xs_3 = FSharpList__get_Tail(xs_3);
    }
    const xs_6: FSharpList<State> = node;
    const t_4: FSharpList<State> = FSharpList_get_Empty();
    xs_6.tail = t_4;
    return FSharpList__get_Tail(root);
}

export function scanBack<T, State>(folder: (arg0: T, arg1: State) => State, xs: FSharpList<T>, state: State): FSharpList<State> {
    return ofArray(scanBack_1(folder, toArray(xs), state, null));
}

export function append<T>(xs: FSharpList<T>, ys: FSharpList<T>): FSharpList<T> {
    return fold((acc: FSharpList<T>, x: T): FSharpList<T> => FSharpList_Cons_305B8EAC(x, acc), ys, reverse(xs));
}

export function collect<T, U>(mapping: (arg0: T) => FSharpList<U>, xs: FSharpList<T>): FSharpList<U> {
    let xs_1: FSharpList<U>, t: FSharpList<U>;
    const root: FSharpList<U> = FSharpList_get_Empty();
    let node: FSharpList<U> = root;
    let ys: FSharpList<T> = xs;
    while (!FSharpList__get_IsEmpty(ys)) {
        let zs: FSharpList<U> = mapping(FSharpList__get_Head(ys));
        while (!FSharpList__get_IsEmpty(zs)) {
            node = ((xs_1 = node, (t = (new FSharpList(FSharpList__get_Head(zs), void 0)), (xs_1.tail = t, t))));
            zs = FSharpList__get_Tail(zs);
        }
        ys = FSharpList__get_Tail(ys);
    }
    const xs_3: FSharpList<U> = node;
    const t_2: FSharpList<U> = FSharpList_get_Empty();
    xs_3.tail = t_2;
    return FSharpList__get_Tail(root);
}

export function mapIndexed<T, U>(mapping: (arg0: int32, arg1: T) => U, xs: FSharpList<T>): FSharpList<U> {
    const root: FSharpList<U> = FSharpList_get_Empty();
    const folder: (arg0: int32, arg1: FSharpList<U>, arg2: T) => FSharpList<U> = (i: int32, acc: FSharpList<U>, x: T): FSharpList<U> => {
        const t: FSharpList<U> = new FSharpList(mapping(i, x), void 0);
        acc.tail = t;
        return t;
    };
    const node: FSharpList<U> = foldIndexed(folder, root, xs);
    const t_2: FSharpList<U> = FSharpList_get_Empty();
    node.tail = t_2;
    return FSharpList__get_Tail(root);
}

export function map<T, U>(mapping: (arg0: T) => U, xs: FSharpList<T>): FSharpList<U> {
    const root: FSharpList<U> = FSharpList_get_Empty();
    const folder: (arg0: FSharpList<U>, arg1: T) => FSharpList<U> = (acc: FSharpList<U>, x: T): FSharpList<U> => {
        const t: FSharpList<U> = new FSharpList(mapping(x), void 0);
        acc.tail = t;
        return t;
    };
    const node: FSharpList<U> = fold(folder, root, xs);
    const t_2: FSharpList<U> = FSharpList_get_Empty();
    node.tail = t_2;
    return FSharpList__get_Tail(root);
}

export function indexed<$a>(xs: FSharpList<$a>): FSharpList<[int32, $a]> {
    return mapIndexed((i: int32, x: $a): [int32, $a] => [i, x], xs);
}

export function map2<T1, T2, U>(mapping: (arg0: T1, arg1: T2) => U, xs: FSharpList<T1>, ys: FSharpList<T2>): FSharpList<U> {
    const root: FSharpList<U> = FSharpList_get_Empty();
    const folder: (arg0: FSharpList<U>, arg1: T1, arg2: T2) => FSharpList<U> = (acc: FSharpList<U>, x: T1, y: T2): FSharpList<U> => {
        const t: FSharpList<U> = new FSharpList(mapping(x, y), void 0);
        acc.tail = t;
        return t;
    };
    const node: FSharpList<U> = fold2(folder, root, xs, ys);
    const t_2: FSharpList<U> = FSharpList_get_Empty();
    node.tail = t_2;
    return FSharpList__get_Tail(root);
}

export function mapIndexed2<T1, T2, U>(mapping: (arg0: int32, arg1: T1, arg2: T2) => U, xs: FSharpList<T1>, ys: FSharpList<T2>): FSharpList<U> {
    const loop: (arg0: int32, arg1: FSharpList<U>, arg2: FSharpList<T1>, arg3: FSharpList<T2>) => FSharpList<U> = (i_mut: int32, acc_mut: FSharpList<U>, xs_1_mut: FSharpList<T1>, ys_1_mut: FSharpList<T2>): FSharpList<U> => {
        let t: FSharpList<U>;
        loop:
        while (true) {
            const i: int32 = i_mut, acc: FSharpList<U> = acc_mut, xs_1: FSharpList<T1> = xs_1_mut, ys_1: FSharpList<T2> = ys_1_mut;
            if (FSharpList__get_IsEmpty(xs_1) ? true : FSharpList__get_IsEmpty(ys_1)) {
                return acc;
            }
            else {
                i_mut = (i + 1);
                acc_mut = ((t = (new FSharpList(mapping(i, FSharpList__get_Head(xs_1), FSharpList__get_Head(ys_1)), void 0)), (acc.tail = t, t)));
                xs_1_mut = FSharpList__get_Tail(xs_1);
                ys_1_mut = FSharpList__get_Tail(ys_1);
                continue loop;
            }
            break;
        }
    };
    const root: FSharpList<U> = FSharpList_get_Empty();
    const node_1: FSharpList<U> = loop(0, root, xs, ys);
    const t_2: FSharpList<U> = FSharpList_get_Empty();
    node_1.tail = t_2;
    return FSharpList__get_Tail(root);
}

export function map3<T1, T2, T3, U>(mapping: (arg0: T1, arg1: T2, arg2: T3) => U, xs: FSharpList<T1>, ys: FSharpList<T2>, zs: FSharpList<T3>): FSharpList<U> {
    const loop: (arg0: FSharpList<U>, arg1: FSharpList<T1>, arg2: FSharpList<T2>, arg3: FSharpList<T3>) => FSharpList<U> = (acc_mut: FSharpList<U>, xs_1_mut: FSharpList<T1>, ys_1_mut: FSharpList<T2>, zs_1_mut: FSharpList<T3>): FSharpList<U> => {
        let t: FSharpList<U>;
        loop:
        while (true) {
            const acc: FSharpList<U> = acc_mut, xs_1: FSharpList<T1> = xs_1_mut, ys_1: FSharpList<T2> = ys_1_mut, zs_1: FSharpList<T3> = zs_1_mut;
            if ((FSharpList__get_IsEmpty(xs_1) ? true : FSharpList__get_IsEmpty(ys_1)) ? true : FSharpList__get_IsEmpty(zs_1)) {
                return acc;
            }
            else {
                acc_mut = ((t = (new FSharpList(mapping(FSharpList__get_Head(xs_1), FSharpList__get_Head(ys_1), FSharpList__get_Head(zs_1)), void 0)), (acc.tail = t, t)));
                xs_1_mut = FSharpList__get_Tail(xs_1);
                ys_1_mut = FSharpList__get_Tail(ys_1);
                zs_1_mut = FSharpList__get_Tail(zs_1);
                continue loop;
            }
            break;
        }
    };
    const root: FSharpList<U> = FSharpList_get_Empty();
    const node_1: FSharpList<U> = loop(root, xs, ys, zs);
    const t_2: FSharpList<U> = FSharpList_get_Empty();
    node_1.tail = t_2;
    return FSharpList__get_Tail(root);
}

export function mapFold<State, T, Result>(mapping: (arg0: State, arg1: T) => [Result, State], state: State, xs: FSharpList<T>): [FSharpList<Result>, State] {
    const folder: (arg0: [FSharpList<Result>, State], arg1: T) => [FSharpList<Result>, State] = (tupledArg: [FSharpList<Result>, State], x: T): [FSharpList<Result>, State] => {
        let t: FSharpList<Result>;
        const patternInput: [Result, State] = mapping(tupledArg[1], x);
        return [(t = (new FSharpList(patternInput[0], void 0)), (tupledArg[0].tail = t, t)), patternInput[1]];
    };
    const root: FSharpList<Result> = FSharpList_get_Empty();
    const patternInput_1: [FSharpList<Result>, State] = fold(folder, [root, state], xs);
    const t_2: FSharpList<Result> = FSharpList_get_Empty();
    patternInput_1[0].tail = t_2;
    return [FSharpList__get_Tail(root), patternInput_1[1]];
}

export function mapFoldBack<T, State, Result>(mapping: (arg0: T, arg1: State) => [Result, State], xs: FSharpList<T>, state: State): [FSharpList<Result>, State] {
    return mapFold((acc: State, x: T): [Result, State] => mapping(x, acc), state, reverse(xs));
}

export function tryPick<T, $a>(f: (arg0: T) => Option<$a>, xs: FSharpList<T>): Option<$a> {
    const loop: (arg0: FSharpList<T>) => Option<$a> = (xs_1_mut: FSharpList<T>): Option<$a> => {
        loop:
        while (true) {
            const xs_1: FSharpList<T> = xs_1_mut;
            if (FSharpList__get_IsEmpty(xs_1)) {
                return void 0;
            }
            else {
                const matchValue: Option<$a> = f(FSharpList__get_Head(xs_1));
                if (matchValue == null) {
                    xs_1_mut = FSharpList__get_Tail(xs_1);
                    continue loop;
                }
                else {
                    return matchValue;
                }
            }
            break;
        }
    };
    return loop(xs);
}

export function pick<$a, $b>(f: (arg0: $a) => Option<$b>, xs: FSharpList<$a>): $b {
    const matchValue: Option<$b> = tryPick(f, xs);
    if (matchValue == null) {
        return indexNotFound();
    }
    else {
        return value_1(matchValue);
    }
}

export function tryFind<$a>(f: (arg0: $a) => boolean, xs: FSharpList<$a>): Option<$a> {
    return tryPick((x: $a): Option<$a> => (f(x) ? some(x) : (void 0)), xs);
}

export function find<$a>(f: (arg0: $a) => boolean, xs: FSharpList<$a>): $a {
    const matchValue: Option<$a> = tryFind(f, xs);
    if (matchValue == null) {
        return indexNotFound();
    }
    else {
        return value_1(matchValue);
    }
}

export function tryFindBack<$a>(f: (arg0: $a) => boolean, xs: FSharpList<$a>): Option<$a> {
    return tryFindBack_1(f, toArray(xs));
}

export function findBack<$a>(f: (arg0: $a) => boolean, xs: FSharpList<$a>): $a {
    const matchValue: Option<$a> = tryFindBack(f, xs);
    if (matchValue == null) {
        return indexNotFound();
    }
    else {
        return value_1(matchValue);
    }
}

export function tryFindIndex<T>(f: (arg0: T) => boolean, xs: FSharpList<T>): Option<int32> {
    const loop: (arg0: int32, arg1: FSharpList<T>) => Option<int32> = (i_mut: int32, xs_1_mut: FSharpList<T>): Option<int32> => {
        loop:
        while (true) {
            const i: int32 = i_mut, xs_1: FSharpList<T> = xs_1_mut;
            if (FSharpList__get_IsEmpty(xs_1)) {
                return void 0;
            }
            else if (f(FSharpList__get_Head(xs_1))) {
                return i;
            }
            else {
                i_mut = (i + 1);
                xs_1_mut = FSharpList__get_Tail(xs_1);
                continue loop;
            }
            break;
        }
    };
    return loop(0, xs);
}

export function findIndex<$a>(f: (arg0: $a) => boolean, xs: FSharpList<$a>): int32 {
    const matchValue: Option<int32> = tryFindIndex(f, xs);
    if (matchValue == null) {
        indexNotFound();
        return -1;
    }
    else {
        return value_1(matchValue) | 0;
    }
}

export function tryFindIndexBack<$a>(f: (arg0: $a) => boolean, xs: FSharpList<$a>): Option<int32> {
    return tryFindIndexBack_1(f, toArray(xs));
}

export function findIndexBack<$a>(f: (arg0: $a) => boolean, xs: FSharpList<$a>): int32 {
    const matchValue: Option<int32> = tryFindIndexBack(f, xs);
    if (matchValue == null) {
        indexNotFound();
        return -1;
    }
    else {
        return value_1(matchValue) | 0;
    }
}

export function tryItem<T>(n: int32, xs: FSharpList<T>): Option<T> {
    const loop: (arg0: int32, arg1: FSharpList<T>) => Option<T> = (i_mut: int32, xs_1_mut: FSharpList<T>): Option<T> => {
        loop:
        while (true) {
            const i: int32 = i_mut, xs_1: FSharpList<T> = xs_1_mut;
            if (FSharpList__get_IsEmpty(xs_1)) {
                return void 0;
            }
            else if (i === n) {
                return some(FSharpList__get_Head(xs_1));
            }
            else {
                i_mut = (i + 1);
                xs_1_mut = FSharpList__get_Tail(xs_1);
                continue loop;
            }
            break;
        }
    };
    return loop(0, xs);
}

export function item<T>(n: int32, xs: FSharpList<T>): T {
    return FSharpList__get_Item_Z524259A4(xs, n);
}

export function filter<T>(f: (arg0: T) => boolean, xs: FSharpList<T>): FSharpList<T> {
    const root: FSharpList<T> = FSharpList_get_Empty();
    const folder: (arg0: FSharpList<T>, arg1: T) => FSharpList<T> = (acc: FSharpList<T>, x: T): FSharpList<T> => {
        if (f(x)) {
            const t: FSharpList<T> = new FSharpList(x, void 0);
            acc.tail = t;
            return t;
        }
        else {
            return acc;
        }
    };
    const node: FSharpList<T> = fold(folder, root, xs);
    const t_2: FSharpList<T> = FSharpList_get_Empty();
    node.tail = t_2;
    return FSharpList__get_Tail(root);
}

export function partition<T>(f: (arg0: T) => boolean, xs: FSharpList<T>): [FSharpList<T>, FSharpList<T>] {
    const patternInput: [FSharpList<T>, FSharpList<T>] = [FSharpList_get_Empty(), FSharpList_get_Empty()];
    const root2: FSharpList<T> = patternInput[1];
    const root1: FSharpList<T> = patternInput[0];
    const folder: (arg0: [FSharpList<T>, FSharpList<T>], arg1: T) => [FSharpList<T>, FSharpList<T>] = (tupledArg: [FSharpList<T>, FSharpList<T>], x: T): [FSharpList<T>, FSharpList<T>] => {
        let t: FSharpList<T>, t_2: FSharpList<T>;
        const lacc: FSharpList<T> = tupledArg[0];
        const racc: FSharpList<T> = tupledArg[1];
        if (f(x)) {
            return [(t = (new FSharpList(x, void 0)), (lacc.tail = t, t)), racc];
        }
        else {
            return [lacc, (t_2 = (new FSharpList(x, void 0)), (racc.tail = t_2, t_2))];
        }
    };
    const patternInput_1: [FSharpList<T>, FSharpList<T>] = fold(folder, [root1, root2], xs);
    const t_4: FSharpList<T> = FSharpList_get_Empty();
    patternInput_1[0].tail = t_4;
    const t_5: FSharpList<T> = FSharpList_get_Empty();
    patternInput_1[1].tail = t_5;
    return [FSharpList__get_Tail(root1), FSharpList__get_Tail(root2)];
}

export function choose<T>(f: (arg0: T) => Option<T>, xs: FSharpList<T>): FSharpList<T> {
    const root: FSharpList<T> = FSharpList_get_Empty();
    const folder: (arg0: FSharpList<T>, arg1: T) => FSharpList<T> = (acc: FSharpList<T>, x: T): FSharpList<T> => {
        const matchValue: Option<T> = f(x);
        if (matchValue == null) {
            return acc;
        }
        else {
            const t: FSharpList<T> = new FSharpList(value_1(matchValue), void 0);
            acc.tail = t;
            return t;
        }
    };
    const node: FSharpList<T> = fold(folder, root, xs);
    const t_2: FSharpList<T> = FSharpList_get_Empty();
    node.tail = t_2;
    return FSharpList__get_Tail(root);
}

export function contains<T>(value: T, xs: FSharpList<T>, eq: IEqualityComparer<T>): boolean {
    return tryFindIndex((v: T): boolean => eq.Equals(value, v), xs) != null;
}

export function initialize<T>(n: int32, f: (arg0: int32) => T): FSharpList<T> {
    let xs: FSharpList<T>, t: FSharpList<T>;
    const root: FSharpList<T> = FSharpList_get_Empty();
    let node: FSharpList<T> = root;
    for (let i: int32 = 0; i <= (n - 1); i++) {
        node = ((xs = node, (t = (new FSharpList(f(i), void 0)), (xs.tail = t, t))));
    }
    const xs_2: FSharpList<T> = node;
    const t_2: FSharpList<T> = FSharpList_get_Empty();
    xs_2.tail = t_2;
    return FSharpList__get_Tail(root);
}

export function replicate<$a>(n: int32, x: $a): FSharpList<$a> {
    return initialize(n, (_arg: int32): $a => x);
}

export function reduce<T>(f: (arg0: T, arg1: T) => T, xs: FSharpList<T>): T {
    if (FSharpList__get_IsEmpty(xs)) {
        throw new Error(SR_inputWasEmpty);
    }
    else {
        return fold(f, head(xs), tail(xs));
    }
}

export function reduceBack<T>(f: (arg0: T, arg1: T) => T, xs: FSharpList<T>): T {
    if (FSharpList__get_IsEmpty(xs)) {
        throw new Error(SR_inputWasEmpty);
    }
    else {
        return foldBack(f, tail(xs), head(xs));
    }
}

export function forAll<$a>(f: (arg0: $a) => boolean, xs: FSharpList<$a>): boolean {
    return fold((acc: boolean, x: $a): boolean => (acc && f(x)), true, xs);
}

export function forAll2<$a, $b>(f: (arg0: $a, arg1: $b) => boolean, xs: FSharpList<$a>, ys: FSharpList<$b>): boolean {
    return fold2((acc: boolean, x: $a, y: $b): boolean => (acc && f(x, y)), true, xs, ys);
}

export function exists<$a>(f: (arg0: $a) => boolean, xs: FSharpList<$a>): boolean {
    return tryFindIndex(f, xs) != null;
}

export function exists2<T1, T2>(f_mut: (arg0: T1, arg1: T2) => boolean, xs_mut: FSharpList<T1>, ys_mut: FSharpList<T2>): boolean {
    exists2:
    while (true) {
        const f: (arg0: T1, arg1: T2) => boolean = f_mut, xs: FSharpList<T1> = xs_mut, ys: FSharpList<T2> = ys_mut;
        const matchValue: boolean = FSharpList__get_IsEmpty(xs);
        const matchValue_1: boolean = FSharpList__get_IsEmpty(ys);
        if (matchValue) {
            if (matchValue_1) {
                return false;
            }
            else {
                throw new Error((SR_differentLengths + "\\nParameter name: ") + "list2");
            }
        }
        else if (matchValue_1) {
            throw new Error((SR_differentLengths + "\\nParameter name: ") + "list2");
        }
        else if (f(FSharpList__get_Head(xs), FSharpList__get_Head(ys))) {
            return true;
        }
        else {
            f_mut = f;
            xs_mut = FSharpList__get_Tail(xs);
            ys_mut = FSharpList__get_Tail(ys);
            continue exists2;
        }
        break;
    }
}

export function unzip<$a, $b>(xs: FSharpList<[$a, $b]>): [FSharpList<$a>, FSharpList<$b>] {
    return foldBack((tupledArg: [$a, $b], tupledArg_1: [FSharpList<$a>, FSharpList<$b>]): [FSharpList<$a>, FSharpList<$b>] => [FSharpList_Cons_305B8EAC(tupledArg[0], tupledArg_1[0]), FSharpList_Cons_305B8EAC(tupledArg[1], tupledArg_1[1])], xs, [FSharpList_get_Empty(), FSharpList_get_Empty()]);
}

export function unzip3<$a, $b, $c>(xs: FSharpList<[$a, $b, $c]>): [FSharpList<$a>, FSharpList<$b>, FSharpList<$c>] {
    return foldBack((tupledArg: [$a, $b, $c], tupledArg_1: [FSharpList<$a>, FSharpList<$b>, FSharpList<$c>]): [FSharpList<$a>, FSharpList<$b>, FSharpList<$c>] => [FSharpList_Cons_305B8EAC(tupledArg[0], tupledArg_1[0]), FSharpList_Cons_305B8EAC(tupledArg[1], tupledArg_1[1]), FSharpList_Cons_305B8EAC(tupledArg[2], tupledArg_1[2])], xs, [FSharpList_get_Empty(), FSharpList_get_Empty(), FSharpList_get_Empty()]);
}

export function zip<$a, $b>(xs: FSharpList<$a>, ys: FSharpList<$b>): FSharpList<[$a, $b]> {
    return map2((x: $a, y: $b): [$a, $b] => [x, y], xs, ys);
}

export function zip3<$a, $b, $c>(xs: FSharpList<$a>, ys: FSharpList<$b>, zs: FSharpList<$c>): FSharpList<[$a, $b, $c]> {
    return map3((x: $a, y: $b, z: $c): [$a, $b, $c] => [x, y, z], xs, ys, zs);
}

export function sortWith<T>(comparer: (arg0: T, arg1: T) => int32, xs: FSharpList<T>): FSharpList<T> {
    const arr: T[] = toArray(xs);
    arr.sort(comparer);
    return ofArray(arr);
}

export function sort<T>(xs: FSharpList<T>, comparer: IComparer<T>): FSharpList<T> {
    return sortWith((x: T, y: T): int32 => comparer.Compare(x, y), xs);
}

export function sortBy<T, U>(projection: (arg0: T) => U, xs: FSharpList<T>, comparer: IComparer<U>): FSharpList<T> {
    return sortWith((x: T, y: T): int32 => comparer.Compare(projection(x), projection(y)), xs);
}

export function sortDescending<T>(xs: FSharpList<T>, comparer: IComparer<T>): FSharpList<T> {
    return sortWith((x: T, y: T): int32 => (comparer.Compare(x, y) * -1), xs);
}

export function sortByDescending<T, U>(projection: (arg0: T) => U, xs: FSharpList<T>, comparer: IComparer<U>): FSharpList<T> {
    return sortWith((x: T, y: T): int32 => (comparer.Compare(projection(x), projection(y)) * -1), xs);
}

export function sum<T>(xs: FSharpList<T>, adder: any): T {
    return fold((acc: T, x: T): T => adder.Add(acc, x), adder.GetZero(), xs);
}

export function sumBy<T, U>(f: (arg0: T) => U, xs: FSharpList<T>, adder: any): U {
    return fold((acc: U, x: T): U => adder.Add(acc, f(x)), adder.GetZero(), xs);
}

export function maxBy<T, U>(projection: (arg0: T) => U, xs: FSharpList<T>, comparer: IComparer<U>): T {
    return reduce((x: T, y: T): T => ((comparer.Compare(projection(y), projection(x)) > 0) ? y : x), xs);
}

export function max<T>(xs: FSharpList<T>, comparer: IComparer<T>): T {
    return reduce((x: T, y: T): T => ((comparer.Compare(y, x) > 0) ? y : x), xs);
}

export function minBy<T, U>(projection: (arg0: T) => U, xs: FSharpList<T>, comparer: IComparer<U>): T {
    return reduce((x: T, y: T): T => ((comparer.Compare(projection(y), projection(x)) > 0) ? x : y), xs);
}

export function min<T>(xs: FSharpList<T>, comparer: IComparer<T>): T {
    return reduce((x: T, y: T): T => ((comparer.Compare(y, x) > 0) ? x : y), xs);
}

export function average<T>(xs: FSharpList<T>, averager: any): T {
    let count: int32 = 0;
    const folder: (arg0: T, arg1: T) => T = (acc: T, x: T): T => {
        count = ((count + 1) | 0);
        return averager.Add(acc, x);
    };
    const total: T = fold(folder, averager.GetZero(), xs);
    return averager.DivideByInt(total, count);
}

export function averageBy<T, U>(f: (arg0: T) => U, xs: FSharpList<T>, averager: any): U {
    let count: int32 = 0;
    const total: U = fold((acc: U, x: T): U => {
        count = ((count + 1) | 0);
        return averager.Add(acc, f(x));
    }, averager.GetZero(), xs);
    return averager.DivideByInt(total, count);
}

export function permute<T>(f: (arg0: int32) => int32, xs: FSharpList<T>): FSharpList<T> {
    return ofArray(permute_1(f, toArray(xs)));
}

export function chunkBySize<T>(chunkSize: int32, xs: FSharpList<T>): FSharpList<FSharpList<T>> {
    return ofArray(map_1(ofArray, chunkBySize_1(chunkSize, toArray(xs)), null));
}

export function allPairs<T1, T2>(xs: FSharpList<T1>, ys: FSharpList<T2>): FSharpList<[T1, T2]> {
    const root: FSharpList<[T1, T2]> = FSharpList_get_Empty();
    let node: FSharpList<[T1, T2]> = root;
    iterate((x: T1): void => {
        iterate((y: T2): void => {
            let xs_1: FSharpList<[T1, T2]>, t: FSharpList<[T1, T2]>;
            node = ((xs_1 = node, (t = (new FSharpList([x, y], void 0)), (xs_1.tail = t, t))));
        }, ys);
    }, xs);
    const xs_3: FSharpList<[T1, T2]> = node;
    const t_2: FSharpList<[T1, T2]> = FSharpList_get_Empty();
    xs_3.tail = t_2;
    return FSharpList__get_Tail(root);
}

export function skip<T>(count_mut: int32, xs_mut: FSharpList<T>): FSharpList<T> {
    skip:
    while (true) {
        const count: int32 = count_mut, xs: FSharpList<T> = xs_mut;
        if (count <= 0) {
            return xs;
        }
        else if (FSharpList__get_IsEmpty(xs)) {
            throw new Error((SR_notEnoughElements + "\\nParameter name: ") + "list");
        }
        else {
            count_mut = (count - 1);
            xs_mut = FSharpList__get_Tail(xs);
            continue skip;
        }
        break;
    }
}

export function skipWhile<T>(predicate_mut: (arg0: T) => boolean, xs_mut: FSharpList<T>): FSharpList<T> {
    skipWhile:
    while (true) {
        const predicate: (arg0: T) => boolean = predicate_mut, xs: FSharpList<T> = xs_mut;
        if (FSharpList__get_IsEmpty(xs)) {
            return xs;
        }
        else if (!predicate(FSharpList__get_Head(xs))) {
            return xs;
        }
        else {
            predicate_mut = predicate;
            xs_mut = FSharpList__get_Tail(xs);
            continue skipWhile;
        }
        break;
    }
}

export function take<T>(count: int32, xs: FSharpList<T>): FSharpList<T> {
    if (count < 0) {
        throw new Error((SR_inputMustBeNonNegative + "\\nParameter name: ") + "count");
    }
    const loop: (arg0: int32, arg1: FSharpList<T>, arg2: FSharpList<T>) => FSharpList<T> = (i_mut: int32, acc_mut: FSharpList<T>, xs_1_mut: FSharpList<T>): FSharpList<T> => {
        let t: FSharpList<T>;
        loop:
        while (true) {
            const i: int32 = i_mut, acc: FSharpList<T> = acc_mut, xs_1: FSharpList<T> = xs_1_mut;
            if (i <= 0) {
                return acc;
            }
            else if (FSharpList__get_IsEmpty(xs_1)) {
                throw new Error((SR_notEnoughElements + "\\nParameter name: ") + "list");
            }
            else {
                i_mut = (i - 1);
                acc_mut = ((t = (new FSharpList(FSharpList__get_Head(xs_1), void 0)), (acc.tail = t, t)));
                xs_1_mut = FSharpList__get_Tail(xs_1);
                continue loop;
            }
            break;
        }
    };
    const root: FSharpList<T> = FSharpList_get_Empty();
    const node: FSharpList<T> = loop(count, root, xs);
    const t_2: FSharpList<T> = FSharpList_get_Empty();
    node.tail = t_2;
    return FSharpList__get_Tail(root);
}

export function takeWhile<T>(predicate: (arg0: T) => boolean, xs: FSharpList<T>): FSharpList<T> {
    const loop: (arg0: FSharpList<T>, arg1: FSharpList<T>) => FSharpList<T> = (acc_mut: FSharpList<T>, xs_1_mut: FSharpList<T>): FSharpList<T> => {
        let t: FSharpList<T>;
        loop:
        while (true) {
            const acc: FSharpList<T> = acc_mut, xs_1: FSharpList<T> = xs_1_mut;
            if (FSharpList__get_IsEmpty(xs_1)) {
                return acc;
            }
            else if (!predicate(FSharpList__get_Head(xs_1))) {
                return acc;
            }
            else {
                acc_mut = ((t = (new FSharpList(FSharpList__get_Head(xs_1), void 0)), (acc.tail = t, t)));
                xs_1_mut = FSharpList__get_Tail(xs_1);
                continue loop;
            }
            break;
        }
    };
    const root: FSharpList<T> = FSharpList_get_Empty();
    const node: FSharpList<T> = loop(root, xs);
    const t_2: FSharpList<T> = FSharpList_get_Empty();
    node.tail = t_2;
    return FSharpList__get_Tail(root);
}

export function truncate<T>(count: int32, xs: FSharpList<T>): FSharpList<T> {
    const loop: (arg0: int32, arg1: FSharpList<T>, arg2: FSharpList<T>) => FSharpList<T> = (i_mut: int32, acc_mut: FSharpList<T>, xs_1_mut: FSharpList<T>): FSharpList<T> => {
        let t: FSharpList<T>;
        loop:
        while (true) {
            const i: int32 = i_mut, acc: FSharpList<T> = acc_mut, xs_1: FSharpList<T> = xs_1_mut;
            if (i <= 0) {
                return acc;
            }
            else if (FSharpList__get_IsEmpty(xs_1)) {
                return acc;
            }
            else {
                i_mut = (i - 1);
                acc_mut = ((t = (new FSharpList(FSharpList__get_Head(xs_1), void 0)), (acc.tail = t, t)));
                xs_1_mut = FSharpList__get_Tail(xs_1);
                continue loop;
            }
            break;
        }
    };
    const root: FSharpList<T> = FSharpList_get_Empty();
    const node: FSharpList<T> = loop(count, root, xs);
    const t_2: FSharpList<T> = FSharpList_get_Empty();
    node.tail = t_2;
    return FSharpList__get_Tail(root);
}

export function getSlice<T>(startIndex: Option<int32>, endIndex: Option<int32>, xs: FSharpList<T>): FSharpList<T> {
    const len: int32 = length(xs) | 0;
    const startIndex_1: int32 = defaultArg(startIndex, 0) | 0;
    const endIndex_1: int32 = defaultArg(endIndex, len - 1) | 0;
    if (startIndex_1 < 0) {
        throw new Error((SR_indexOutOfBounds + "\\nParameter name: ") + "startIndex");
    }
    else if (endIndex_1 >= len) {
        throw new Error((SR_indexOutOfBounds + "\\nParameter name: ") + "endIndex");
    }
    else if (endIndex_1 < startIndex_1) {
        return FSharpList_get_Empty();
    }
    else {
        return take((endIndex_1 - startIndex_1) + 1, skip(startIndex_1, xs));
    }
}

export function splitAt<T>(index: int32, xs: FSharpList<T>): [FSharpList<T>, FSharpList<T>] {
    if (index < 0) {
        throw new Error((SR_inputMustBeNonNegative + "\\nParameter name: ") + "index");
    }
    if (index > FSharpList__get_Length(xs)) {
        throw new Error((SR_notEnoughElements + "\\nParameter name: ") + "index");
    }
    return [take(index, xs), skip(index, xs)];
}

export function exactlyOne<T>(xs: FSharpList<T>): T {
    if (FSharpList__get_IsEmpty(xs)) {
        throw new Error((SR_inputSequenceEmpty + "\\nParameter name: ") + "list");
    }
    else if (FSharpList__get_IsEmpty(FSharpList__get_Tail(xs))) {
        return FSharpList__get_Head(xs);
    }
    else {
        throw new Error((SR_inputSequenceTooLong + "\\nParameter name: ") + "list");
    }
}

export function tryExactlyOne<T>(xs: FSharpList<T>): Option<T> {
    if ((!FSharpList__get_IsEmpty(xs)) && FSharpList__get_IsEmpty(FSharpList__get_Tail(xs))) {
        return some(FSharpList__get_Head(xs));
    }
    else {
        return void 0;
    }
}

export function where<T>(predicate: (arg0: T) => boolean, xs: FSharpList<T>): FSharpList<T> {
    return filter(predicate, xs);
}

export function pairwise<T>(xs: FSharpList<T>): FSharpList<[T, T]> {
    return ofArray(pairwise_1(toArray(xs)));
}

export function windowed<T>(windowSize: int32, xs: FSharpList<T>): FSharpList<FSharpList<T>> {
    return ofArray(map_1(ofArray, windowed_1(windowSize, toArray(xs)), null));
}

export function splitInto<T>(chunks: int32, xs: FSharpList<T>): FSharpList<FSharpList<T>> {
    return ofArray(map_1(ofArray, splitInto_1(chunks, toArray(xs)), null));
}

export function transpose<T>(lists: Iterable<FSharpList<T>>): FSharpList<FSharpList<T>> {
    return ofArray(map_1(ofArray, transpose_1(map_1(toArray, Array.from(lists), null), null), null));
}

export function insertAt<T>(index: int32, y: T, xs: FSharpList<T>): FSharpList<T> {
    let i: int32 = -1;
    let isDone: boolean = false;
    const result: FSharpList<T> = fold((acc: FSharpList<T>, x: T): FSharpList<T> => {
        i = ((i + 1) | 0);
        if (i === index) {
            isDone = true;
            return FSharpList_Cons_305B8EAC(x, FSharpList_Cons_305B8EAC(y, acc));
        }
        else {
            return FSharpList_Cons_305B8EAC(x, acc);
        }
    }, FSharpList_get_Empty(), xs);
    return reverse(isDone ? result : (((i + 1) === index) ? FSharpList_Cons_305B8EAC(y, result) : (() => {
        throw new Error((SR_indexOutOfBounds + "\\nParameter name: ") + "index");
    })()));
}

export function insertManyAt<T>(index: int32, ys: Iterable<T>, xs: FSharpList<T>): FSharpList<T> {
    let i: int32 = -1;
    let isDone: boolean = false;
    const ys_1: FSharpList<T> = ofSeq(ys);
    const result: FSharpList<T> = fold((acc: FSharpList<T>, x: T): FSharpList<T> => {
        i = ((i + 1) | 0);
        if (i === index) {
            isDone = true;
            return FSharpList_Cons_305B8EAC(x, append(ys_1, acc));
        }
        else {
            return FSharpList_Cons_305B8EAC(x, acc);
        }
    }, FSharpList_get_Empty(), xs);
    return reverse(isDone ? result : (((i + 1) === index) ? append(ys_1, result) : (() => {
        throw new Error((SR_indexOutOfBounds + "\\nParameter name: ") + "index");
    })()));
}

export function removeAt<T>(index: int32, xs: FSharpList<T>): FSharpList<T> {
    let i: int32 = -1;
    let isDone: boolean = false;
    const ys: FSharpList<T> = filter((_arg: T): boolean => {
        i = ((i + 1) | 0);
        if (i === index) {
            isDone = true;
            return false;
        }
        else {
            return true;
        }
    }, xs);
    if (!isDone) {
        throw new Error((SR_indexOutOfBounds + "\\nParameter name: ") + "index");
    }
    return ys;
}

export function removeManyAt<T>(index: int32, count: int32, xs: FSharpList<T>): FSharpList<T> {
    let i: int32 = -1;
    let status: int32 = -1;
    const ys: FSharpList<T> = filter((_arg: T): boolean => {
        i = ((i + 1) | 0);
        if (i === index) {
            status = 0;
            return false;
        }
        else if (i > index) {
            if (i < (index + count)) {
                return false;
            }
            else {
                status = 1;
                return true;
            }
        }
        else {
            return true;
        }
    }, xs);
    const status_1: int32 = (((status === 0) && ((i + 1) === (index + count))) ? 1 : status) | 0;
    if (status_1 < 1) {
        throw new Error((SR_indexOutOfBounds + "\\nParameter name: ") + ((status_1 < 0) ? "index" : "count"));
    }
    return ys;
}

export function updateAt<T>(index: int32, y: T, xs: FSharpList<T>): FSharpList<T> {
    let isDone: boolean = false;
    const ys: FSharpList<T> = mapIndexed((i: int32, x: T): T => {
        if (i === index) {
            isDone = true;
            return y;
        }
        else {
            return x;
        }
    }, xs);
    if (!isDone) {
        throw new Error((SR_indexOutOfBounds + "\\nParameter name: ") + "index");
    }
    return ys;
}

